"""
Functions and classes for handling next generation sequencing data.
:author: Manuel Tuschen
:license: FreeBSD

License
----------
Copyright (c) 2016, Manuel Tuschen
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS I" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import division, absolute_import, unicode_literals, print_function

from .ngsBED import BED, BEDEntry
from .ngsBedGraph import BedGraph, BedGraphEntry




def fromGFF2BED(inGFF):
    """
    Convert data from GFF to BED. As both formats are not equivalent only
    partial conversion can be performed. All track specific entries are not
    converted. If the name field should be set, an attribute ID must be
    specified. A score can only be set if a name was given. The same is true for
    the strand.

    Parameters
    ----------
    inGFF : GFF
        The GFF object to convert.

    Return
    ------
    outBED : BED
        The converted BED object.
    """
    outBED = BED()
    for gff in inGFF:
        chrom = gff.seqid
        chromStart = gff.start-1  # gff format is 1 based and end inclusive
        chromEnd = gff.end

        bed = BEDEntry(chrom, chromStart, chromEnd)
        if gff.attribues is not None:
            if "ID" in gff.attributes:
                bed.name = gff.attributes["ID"]
                if gff.score is not None:
                    bed.score = gff.score
                    if gff.strand == '+' or gff.strand == '-':
                        bed.strand = gff.strand
        if not bed.valid:
            raise ValueError("Conversion failed due to invalid BED entry")
        outBED.append(bed)
    return outBED


def fromBED2BedGraph(inBED):
    """
    Convert data from  BED to BedGraph.

    Parameters
    ----------
    inBED : BED
        The BED object to convert.

    Return
    ------
    outBedGraph : BedGraph
        The converted BedGraph object.
    """

    outBedGraph = BedGraph()
    for bed in inBED:
        if bed.score is None:
            raise ValueError("Conversion failed due to missing score in BED entry")
        chrom = bed.chrom
        chromStart = bed.chromStart
        chromEnd = bed.chromEnd
        score = bed.score
        bedgr = BedGraphEntry(chrom, chromStart, chromEnd, score)
        if not bedgr.valid:
            raise ValueError("Conversion failed due to invalid BedGraph entry")
        outBedGraph.append(bedgr)
    return outBedGraph


def fromBedGraph2BED(inBedGraph):
    """
    Convert data from BedGraph to  BED. As in BedGraph there is no name filed
    but is required for the score entry the score will not be converted.

    Parameters
    ----------
    inBedGraph : BedGraph
        The BedGraph object to convert.

    Return
    ------
    outBED : BED
        The converted BED object.
    """

    outBED = BED()
    for bedgr in inBedGraph:
        chrom = bedgr.chrom
        chromStart = bedgr.start
        chromEnd = bedgr.end

        bed = BEDEntry(chrom, chromStart, chromEnd)
        if not bed.valid:
            raise ValueError("Conversion failed due to invalid BED entry")
        outBED.append(bed)
    return outBED


def fromWIG2BedGraph(inWIG):
    """
    Convert data from WIG to BedGraph.

    Parameters
    ----------
    inWIG : WIG
        The WIG object to convert.

    Return
    ------
    outBED : BED
        The converted BED object.
    """

    outBedGraph = BedGraph
    for bedgr in inWIG:
        if not bedgr.valid:
            raise ValueError("Conversion failed due to invalid BedGraph entry")
        outBedGraph.append(bedgr)
    return outBedGraph


"""
Functions and classes for handling next generation sequencing data.
:author: Manuel Tuschen
:license: FreeBSD

License
----------
Copyright (c) 2016, Manuel Tuschen
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


from __future__ import division, absolute_import, unicode_literals, print_function
import re

from miscpy import prepareLoading

from ._ngsData import Data
from .ngsBED import BED6Entry


__all__ = ["BroadPeakEntry", "NarrowPeakEntry", "BroadPeak", "NarrowPeak"]


class BroadPeakEntry(BED6Entry):
    """
    This format is used to provide called regions of signal enrichment based on
    pooled, normalized (interpreted) data. It is a BED 6+3 format:
    https://genome.ucsc.edu/FAQ/FAQformat.html#format13

 Attributes
    ----------
    chrom : basestring
        Name of the chromosome (or contig, scaffold, etc.).
    chromStart : int
       The starting position of the feature in the chromosome or scaffold. The
       first base in a chromosome is numbered 0.
    chromEnd : int
        The ending position of the feature in the chromosome or scaffold. The
        chromEnd base is not included in the display of the feature. For
        example, the first 100 bases of a chromosome are defined as
        chromStart=0, chromEnd=100, and span the bases numbered 0-99.
    name : basestring
        Name given to a region (preferably unique). Use '.' if no name is
        assigned.
    score : int
        Indicates how dark the peak will be displayed in the browser (0-1000)
    strand : char
         +/- to denote strand or orientation (whenever applicable). Use '.' if
         no orientation is assigned.
    signalValue: float
        Measurement of overall (usually, average) enrichment for the region.
    pValue : float
        Measurement of statistical significance (-log10). Use -1 if no pValue
        is assigned.
    qValue : float
        Measurement of statistical significance using false discovery rate
        (-log10). Use -1 if no qValue is assigned.
    """

    def __init__(self, chrom, chromStart, chromEnd, score, signalValue,
                 name=None, strand=None, pValue=None, qValue=None):

        super().__init__(chrom, chromStart, chromEnd, score, name, strand)
        self.signalValue = signalValue
        self.pValue = pValue
        self.qValue = qValue

    def __str__(self):
        sep = self._sep
        newline = self._newline

        signalValue = str(self.signalValue)
        if self.pValue is None:
            pValue = str(-1)
        else:
            pValue = str(self.pValue)
        if self.qValue is None:
            qValue = str(-1)
        else:
            qValue = str(self.qValue)
        return super().__str__().rstrip() + sep + signalValue + sep + pValue + sep + qValue + newline

    def __eq__(self, other):

        if super().__eq__(
                other) and self.signalValue == other.signalValue and self.pValue == other.pValue and self.qValue == other.qValue:
            return True
        else:
            return False

    def _valid(self):

        if not super()._valid():
            return False

        # check signal Value
        if not isinstance(self.signalValue, float):
            return False
        # check pValue
        if not (self.pValue is None or isinstance(self.pValue,
                                                  float) or self.pValue == -1):
            return False
        # check qValue
        if not (self.qValue is None or isinstance(self.qValue,
                                                  float) or self.qValue == -1):
            return False
        return True


class NarrowPeakEntry(BroadPeakEntry):
    """
    This format is used to provide called regions of signal enrichment based on
    pooled, normalized (interpreted) data. It is a BED 6+3 format:
    https://genome.ucsc.edu/FAQ/FAQformat.html#format12

 Attributes
    ----------
    chrom : basestring
        Name of the chromosome (or contig, scaffold, etc.).
    chromStart : int
       The starting position of the feature in the chromosome or scaffold. The
       first base in a chromosome is numbered 0.
    chromEnd : int
        The ending position of the feature in the chromosome or scaffold. The
        chromEnd base is not included in the display of the feature. For
        example, the first 100 bases of a chromosome are defined as
        chromStart=0, chromEnd=100, and span the bases numbered 0-99.
    name : basestring
        Name given to a region (preferably unique). Use '.' if no name is
        assigned.
    score : int
        Indicates how dark the peak will be displayed in the browser (0-1000)
    strand : char
         +/- to denote strand or orientation (whenever applicable). Use '.' if
         no orientation is assigned.
    signalValue: float
        Measurement of overall (usually, average) enrichment for the region.
    pValue : float
        Measurement of statistical significance (-log10). Use -1 if no pValue
        is assigned.
    qValue : float
        Measurement of statistical significance using false discovery rate
        (-log10). Use -1 if no qValue is assigned.
    peak : int
        Point-source called for this peak; 0-based offset from chromStart. Use -1 if no point-source called
    """

    def __init__(self, chrom, chromStart, chromEnd, score, signalValue,
                 name=None, strand=None, pValue=None, qValue=None, peak=None):

        super().__init__(chrom, chromStart, chromEnd, score, signalValue, name,
                         strand, pValue, qValue)
        self.peak = peak

    def __str__(self):
        sep = self._sep
        newline = self._newline

        if self.peak is None:
            peak = str(-1)
        else:
            peak = str(self.peak)

        return super().__str__().rstrip() + sep + peak + newline

    def __eq__(self, other):

        if super().__eq__(other) and self.peak == other.peak:
            return True
        else:
            return False

    def _valid(self):
        if not super()._valid():
            return False
        # check peak Value
        if not (self.qValue is None or isinstance(self.qValue,
                                                  int) or self.qValue == -1):
            return False
        return True


class RNAelementsEntry(BED6Entry):
    """
    This fromat provides fields for RNAseq data. It is a BED 6+3 format:
    https://genome.ucsc.edu/FAQ/FAQformat.html#format11

 Attributes
    ----------
    chrom : basestring
        Name of the chromosome (or contig, scaffold, etc.).
    chromStart : int
       The starting position of the feature in the chromosome or scaffold. The
       first base in a chromosome is numbered 0.
    chromEnd : int
        The ending position of the feature in the chromosome or scaffold. The
        chromEnd base is not included in the display of the feature. For
        example, the first 100 bases of a chromosome are defined as
        chromStart=0, chromEnd=100, and span the bases numbered 0-99.
    name : basestring
        Name given to a region (preferably unique). Use '.' if no name is
        assigned.
    score : int
        Indicates how dark the peak will be displayed in the browser (0-1000)
    strand : char
         +/- to denote strand or orientation (whenever applicable). Use '.' if
         no orientation is assigned.
    level: float
        Expression level such as RPKM or FPKM.
    signf : float
        Statistical significance such as IDR.
    score2 : float
        Additional measurement/count e.g. number of reads.
    """

    def __init__(self, chrom, chromStart, chromEnd, score, level, signif, score2,
                 name=None, strand=None):
        super().__init__(chrom, chromStart, chromEnd, score, name, strand)
        self.level = level
        self.sifnif = signif
        self.score2 = score2

    def __str__(self):
        sep = self._sep
        newline = self._newline

        level = str(self.level)
        signif = str(self.signif)
        score2 = str(self.score2)

        return super().__str__().rstrip() + sep + level + sep + signif + sep + score2 + newline

    def __eq__(self, other):
        if super().__eq__(other) and self.level == other.level and self.sifnif == other.signif and self.score == other.score:
            return True
        else:
            return False

    def _valid(self):
        if not super().__valid():
            return False

        # check level
        if not isinstance(self.level, float):
            return False
        # check signif
        if not isinstance(self.signif, float):
            return False
        # check score2
        if not isinstance(self.score2, float):
            return False

        return True


class longRNAexpressionEntry(BED6Entry):
    """
    This fromqat provides fields for RNAseq data. It is a BED 6+4 format:
    https://genome.ucsc.edu/FAQ/FAQformat.html#format11

 Attributes
    ----------
    chrom : basestring
        Name of the chromosome (or contig, scaffold, etc.).
    chromStart : int
       The starting position of the feature in the chromosome or scaffold. The
       first base in a chromosome is numbered 0.
    chromEnd : int
        The ending position of the feature in the chromosome or scaffold. The
        chromEnd base is not included in the display of the feature. For
        example, the first 100 bases of a chromosome are defined as
        chromStart=0, chromEnd=100, and span the bases numbered 0-99.
    name : basestring
        Name given to a region (preferably unique). Use '.' if no name is
        assigned.
    score : int
        Indicates how dark the peak will be displayed in the browser (0-1000)
    strand : char
         +/- to denote strand or orientation (whenever applicable). Use '.' if
         no orientation is assigned.
    length: int
        Length of the transcript
    gene_id : basestring
        Gene ID like Ensembl
    coverage : float
        Coverage
    FPKM : float
        FPKM value
    """

    def __init__(self, chrom, chromStart, chromEnd, score, length, gene_id,
                 coverage, fpkm, name=None, strand=None):
        super().__init__(chrom, chromStart, chromEnd, score, name, strand)
        self.length = length
        self.gene_id = gene_id
        self.coverage = coverage
        self.fpkm = fpkm

    def __str__(self):
        sep = self._sep
        newline = self._newline

        length = str(self.length)
        gene_id = str(self.gene_id)
        coverage = str(self.coverage)
        fpkm = str(self.fpkm)

        return super().__str__().rstrip() + sep + length + sep + gene_id + sep + coverage + sep + fpkm + newline

    def __eq__(self, other):
        if super().__eq__(
                other) and self.length == other.length and self.gene_id == other.gene_id and self.coverage == other.coverage and self.fpkm == other.fpkm:
            return True
        else:
            return False

    def _valid(self):
        pattern = re.compile(".")  # we can have an arbitrary string

        if not super()._valid():
            return False

        # check length
        if not isinstance(self.length, int):
            return False
        # check gene_id
        if not isinstance(self.signif, float):
            return False
        # check score2
        if not isinstance(self.score2, float):
            return False

        return True


class ngsBroadPeak(Data):
    """
    A BroadPeak object is designed to parse data from an BroadPeak file or
    write its data to an BroadPeak file. BroadPeak supports iterator and
    mutable sequence protocols. Items can be accessed by an tuple key (chr, n),
    i.e. the nth entry of chr.

    Attributes
    ----------

    Methods
    -------
    append(value) :
        Append another BroadPeakEntry.
    sort() :
        Sort entries;
    parse(name, path, extension="broadPeak", validate=True) :
        Parse data from path/name.extension and store them in the BroadPeak
        object.
    write(name, path, extension="broadPeak", validate=True) :
        Write data to path/name.extension
    """

    def __init__(self):
        super().__init__(BroadPeakEntry)
        self._extension = "broadPeak"

    def parse(self, name, path, extension="broadPeak", validate=True):
        """
        parse BroadPeak file.
        """
        fname = prepareLoading(name, path, extension=extension)

        splitter = '\t'
        counter = 0
        with open(fname, 'r') as file:
            self._data = {}

            for line in file:
                counter += 1
                # handel header files
                if line[0] in self._header_char:
                    continue

                # find chrom
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BroadPeak file! No 'chrom' entry in line {c}".format(
                            c=counter))
                chrom = line[0:split]
                line = line[split + 1:]
                line = line.lstrip()

                # find chromStart
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BroadPeak file! No 'chromStart' entry in line {c}".format(
                            c=counter))
                chromStart = int(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find chromEnd
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BroadPeak file! No 'chromEnd' entry in line {c}".format(
                            c=counter))
                else:
                    chromEnd = int(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find name
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BroadPeak file! No 'name' entry in line {c}".format(
                            c=counter))
                else:
                    name = line[0:split]
                line = line[split + 1:]
                line = line.lstrip()

                # find score
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BroadPeak file! No 'score' entry in line {c}".format(
                            c=counter))
                else:
                    score = int(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find strand
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BroadPeak file! No 'strand' entry in line {c}".format(
                            c=counter))
                else:
                    strand = line[0:split]
                line = line[split + 1:]
                line = line.lstrip()

                # find signal value
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BroadPeak file! No 'signalValue' entry in line {c}".format(
                            c=counter))
                else:
                    signalValue = float(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find  p value
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BroadPeak file! No 'pValue' entry in line {c}".format(
                            c=counter))
                else:
                    pValue = float(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find  q value
                split = line.find(splitter)
                if split != -1:
                    raise IOError(
                        "Invalid BroadPeak file! 'qValue' entry should be the last entry in line {c}".format(
                            c=counter))
                else:
                    line = line.lstrip()
                    line = line.rstrip()
                    qValue = float(line)

                entry = BroadPeakEntry(chrom, chromStart, chromEnd, score,
                                       signalValue, name, strand, pValue,
                                       qValue)
                if validate:
                    if not entry.valid:
                        raise IOError(
                            "Invalid BroadPeak file! Entry in line {c} is {e}".format(
                                c=counter, e=str(entry)))

                self.append(entry)


class ngsNarrowPeak(Data):
    """
    A NarrowPeak object is designed to parse data from an BroadPeak file or
    write its data to an NarrowPeak file. NarrowPeak supports iterator and
    mutable sequence protocols. Items can be accessed by an tuple key (chr, n),
    i.e. the nth entry of chr.

    Attributes
    ----------

    Methods
    -------
    append(value) :
        Append another NarrowPeakEntry.
    sort() :
        Sort entries;
    parse(name, path, extension="narrowPeak", validate=True) :
        Parse data from path/name.extension and store them in the BroadPeak
        object.
    write(name, path, extension="narrowPeak", validate=True) :
        Write data to path/name.extension
    """

    def __init__(self):
        super().__init__(NarrowPeakEntry)
        self._extension = "narrowPeak"

    def parse(self, name, path, extension="narrowPeak", validate=True):
        """
        parse BroadPeak file.
        """
        fname = prepareLoading(name, path, extension=extension)

        splitter = '\t'
        counter = 0
        with open(fname, 'r') as file:
            self._data = {}

            for line in file:
                counter += 1
                # handel header files
                if line[0] in self._header_char:
                    continue

                # find chrom
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid NarrowPeak file! No 'chrom' entry in line {c}".format(
                            c=counter))
                chrom = line[0:split]
                line = line[split + 1:]
                line = line.lstrip()

                # find chromStart
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid NarrowPeak file! No 'chromStart' entry in line {c}".format(
                            c=counter))
                chromStart = int(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find chromEnd
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid NarrowPeak file! No 'chromEnd' entry in line {c}".format(
                            c=counter))
                else:
                    chromEnd = int(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find name
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid NarrowPeak file! No 'name' entry in line {c}".format(
                            c=counter))
                else:
                    name = line[0:split]
                line = line[split + 1:]
                line = line.lstrip()

                # find score
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid NarrowPeak file! No 'score' entry in line {c}".format(
                            c=counter))
                else:
                    score = int(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find strand
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid NarrowPeak file! No 'strand' entry in line {c}".format(
                            c=counter))
                else:
                    strand = line[0:split]
                line = line[split + 1:]
                line = line.lstrip()

                # find signal value
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid NarrowPeak file! No 'signalValue' entry in line {c}".format(
                            c=counter))
                else:
                    signalValue = float(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find  p value
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid NarrowPeak file! No 'pValue' entry in line {c}".format(
                            c=counter))
                else:
                    pValue = float(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find  q value
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid NarrowPeak file! No 'qValue' entry in line {c}".format(
                            c=counter))
                else:
                    qValue = float(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find peak
                split = line.find(splitter)
                if split != -1:
                    raise IOError(
                        "Invalid NarrowPeak file! 'peak' entry should be the last entry in line {c}".format(
                            c=counter))
                else:
                    line = line.lstrip()
                    line = line.rstrip()
                    peak = int(line)

                entry = NarrowPeakEntry(chrom, chromStart, chromEnd, score,
                                        signalValue, name, strand, pValue,
                                        qValue, peak)
                if validate:
                    if not entry.valid:
                        raise IOError(
                            "Invalid NarrowPeak file! Entry in line {c} is {e}".format(
                                c=counter, e=str(entry)))

                self.append(entry)



"""
Functions and classes for handling next generation sequencing data.
:author: Manuel Tuschen
:license: FreeBSD

License
----------
Copyright (c) 2016, Manuel Tuschen
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import division, absolute_import, unicode_literals, print_function

import re

from miscpy import prepareLoading, prepareSaving

from ._ngsData import Data


__all__ = ["GFF3Entry", "GFF"]

class GFF3Entry:
    """
    GFF version 3 entry.
    http://gmod.org/wiki/GFF3

    Notes on comparing GFF entries.
    Entris are considered equal only if all entris are equal.
    One entry is regarded as larger only if the sequence id is larger or equal
    and the start position is truly larger. One entry is regarded as
    larger-equalonly if the sequence id is larger or equal and the start
    position is larger or equal. The same rules apply for smaller.

    Attributes
    ----------
    seqid : basestring
        The ID of the landmark used to establish the coordinate system for the
        current feature. IDs may contain any characters, but must escape any
        characters not in the set [a-zA-Z0-9.:^*$@!+_?-|]. In particular, IDs
        may not contain unescaped whitespace and must not begin with an
        unescaped ">". To escape a character in this, or any of the other GFF3
        fields, replace it with the percent sign followed by its hexadecimal
        representation. For example, ">" becomes "%E3". See URL Encoding (or:
        'What are those "%20" codes in URLs?') for details
        (http://www.blooberry.com/indexdot/html/topics/urlencoding.htm).
    source : basestring
       The source is a free text qualifier intended to describe the algorithm or
       operating procedure that generated this feature. Typically this is the
       name of a piece of software, such as "Genescan" or a database name, such
       as "Genbank." In effect, the source is used to extend the feature
       ontology by adding a qualifier to the type creating a new composite type
       that is a subclass of the type in the type column. It is not necessary to
       specify a source. If there is no source, put a "." (a period) in this
       field.
    type : basestring
        The type of the feature (previously called the "method"). This is
        constrained to be either: (a) a term from the "lite" sequence ontology,
        SOFA; or (b) a SOFA accession number. The latter alternative is
        distinguished using the syntax SO:000000. This field is required.
    start : int
        The start of the feature, in 1-based integer coordinates, relative to
        the landmark given in column 1. Start is always less than or equal to
        end. For zero-length features, such as insertion sites, start equals
        end and the implied site is to the right of the indicated base in the
        direction of the landmark. The field is required.
    end : int
        The end of the feature, in 1-based integer coordinates, relative to
        the landmark given in column 1. Start is always less than or equal to
        end. For zero-length features, such as insertion sites, start equals
        end and the implied site is to the right of the indicated base in the
        direction of the landmark. The field is required.
    score : float
        The score of the feature, a floating point number. As in earlier
        versions of the format, the semantics of the score are ill-defined.
        It is strongly recommended that E-values be used for sequence
        similarity features, and that P-values be used for ab initio gene
        prediction features. If there is no score, put a "." (a period) in this
        field.
    strand : char
        The strand of the feature. + for positive strand
        (relative to the landmark), - for minus strand, and . for features that
        are not stranded. In addition, ? can be used for features whose
        strandedness is relevant, but unknown.
    phase : int
        For features of type "CDS", the phase indicates where the feature
        begins with reference to the reading frame. The phase is one of the
        integers 0, 1, or 2, indicating the number of bases that should be
        removed from the beginning of this feature to reach the first base of
        the next codon. In other words, a phase of "0" indicates that the next
        codon begins at the first base of the region described by the current
        line, a phase of "1" indicates that the next codon begins at the second
        base of this region, and a phase of "2" indicates that the codon begins
        at the third base of this region. This is NOT to be confused with the
        frame, which is simply start modulo 3. If there is no phase, put a "."
        (a period) in this field. For forward strand features, phase is counted
        from the start field. For reverse strand features, phase is counted
        from the end field. The phase is required for all CDS features.
    attributes : dict
        A list of feature attributes in the format tag=value. Multiple
        tag=value pairs are separated by semicolons. URL escaping rules are
        used for tags or values containing the following characters: ",=;".
        Spaces are allowed in this field, but tabs must be replaced with the
        %09 URL escape. This field is not required.  Multiple attributes of
        the same type are indicated by separating the values with the comma ","
        character.
        Some keys have predefined meanings:
        ID
            Indicates the unique identifier of the feature. IDs must be unique
            within the scope of the GFF file.
        Name
            Display name for the feature. This is the name to be displayed to
            the user. Unlike IDs, there is no requirement that the Name be
            unique within the file.
        Alias
            A secondary name for the feature. It is suggested that this tag be
            used whenever a secondary identifier for the feature is needed,
            such as locus names and accession numbers. Unlike ID, there is no
            requirement that Alias be unique within the file.
        Parent
            Indicates the parent of the feature. A parent ID can be used to
            group exons into transcripts, transcripts into genes, and so forth.
            A feature may have multiple parents. Parent can *only* be used to
            indicate a partof relationship.
        Target
            Indicates the target of a nucleotide-to-nucleotide or
            protein-to-nucleotide alignment. The format of the value is
            "target_id start end [strand]", where strand is optional and may
            be "+" or "-". If the target_id contains spaces, they must be
            escaped as hex escape %20.
        Gap
            The alignment of the feature to the target if the two are not
            collinear (e.g. contain gaps). The alignment format is taken from
            the CIGAR format described in the Exonerate documentation.
            http://cvsweb.sanger.ac.uk/cgi-bin/cvsweb.cgi/exonerate?cvsroot=Ensembl).
            See the GFF3 specification for more information.
        Derives_from
            Used to disambiguate the relationship between one feature and
            another when the relationship is a temporal one rather than a
            purely structural "part of" one. This is needed for polycistronic
            genes. See the GFF3 specification for more information.
        Note
            A free text note.
        Dbxref
            A database cross reference. See the GFF3 specification for more
            information.
        Ontology_term
            A cross reference to an ontology term. See the GFF3 specification
            for more information.
    valid : bool
        To check if the entry is a valid one.

    Methods
    -------
    """

    def __init__(self, seqid, scource, type, start, end, score=None,
                 strand=None, phase=None, attributes=None):

        self.seqid = seqid
        self.source = scource
        self.type = type
        self.start = start
        self.end = end
        self.score = score
        self.strand = strand
        self.phase = phase
        self.attributes = attributes

    def __getattr__(self, valid):
        return self._valid()

    def __str__(self):
        sep = '\t'
        gsep = ';'
        newline = '\n'

        seqid = str(self.seqid)
        source = str(self.source)
        type = str(self.type)
        start = str(self.start)
        end = str(self.end)
        if self.score is None:
            score = '.'
        else:
            score = str(self.score)
        if self.strand is None:
            strand = '.'
        else:
            strand = str(self.strand)
        if self.phase is None:
            phase = '.'
        else:
            phase = str(self.phase)
        group = ""
        for key, value in self.attributes.items():
            group += str(key) + '=' + str(value) + gsep
        group = group.rstrip(gsep)

        return seqid + sep + source + sep + type + sep + start + sep + end + sep + \
               score + sep + strand + sep + phase + sep + group + \
               newline

    def __lt__(self, other):
        if self.seqid <= other.seqid and self.start < other.start:
            return True
        else:
            return False

    def __le__(self, other):
        return not other < self

    def __gt__(self, other):
        return other < self

    def __ge__(self, other):
        return not self < other

    def __eq__(self, other):
        attr_equal = True
        if self.attributes.keys != other.attribues.keys:
            return False
        for key in self.attributes.keys():
            if self.attributes[key] != other.attribues[key]:
                return False

        if self.seqid == other.seqid and self.source == other.source and \
                        self.type == other.type and self.start == other.start and self.end \
                == other.end and self.score == other.score and self.strand == \
                other.starnd and self.pahse == other.phase and attr_equal:
            return True
        else:
            return False

    def __neg__(self, other):
        return not self == other

    def _valid(self):

        pattern = re.compile("[a-zA-Z0-9.:^*$@!+_?-|]")
        strand_pattern = re.compile("[+-.?]")
        # check seqid
        if not pattern.match(self.seqid):
            print("seqid", self.seqid)
            return False

        # check source:
        if not pattern.match(self.source):
            print("source", self.source)
            return False

        # check type:
        if not pattern.match(self.type):
            print("type", self.type)
            return False

        # check start"
        if not isinstance(self.start, int):
            print("start", self.start)
            return False
        if self.start < 1:
            print("start", self.start)
            return False

        # check end
        if not isinstance(self.end, int):
            print("end", self.end)
            return False
        if self.end < self.start:
            print("end", self.end)
            return False

        # check score
        if not (self.score is None or isinstance(self.score, float)):
            print("score", self.score)
            return False

        # check strand
        if not (self.strand is None or strand_pattern.match(self.strand)):
            print("strand", self.strand)
            return False

        # check phase:
        if not (self.phase is None or self.phase in [-1, 0, 1, 2, ]):
            print("phase", self.phase)
            return False

        return True


class GFF(Data):
    """
    A GFF object is designed to parse data from an GFF file or write its
    data to an GFF file. GFF version 3 is currently supported. GFF supports
    iterator and mutable sequence protocols. Items can be accessed by an tuple
    key (seqid, n), i.e. the nth entry of seqid.

    Attributes
    ----------
    version : int
        The GFF version

    Methods
    -------
    append(value) :
        Append another GFFentry.
    sort() :
        Sort entries;
    parse(name, path, extension="gff", validate=True) :
        Parse data from path/name.extension and store them in the GFF object.
    write(name, path, extension="gff", validate=True) :
        Write data to path/name.extension
    """

    def __init__(self, version=3):

        super().__init__(GFF3Entry)
        if version != 3:
            raise ValueError("Requested version not supported")
        self.version = version
        self._header_general = ["##gff-version {v} \n".format(v=self.version)]
        self._header_region = {}
        self._extension = "gff"

    def parse(self, name, path, extension="gff", validate=True):
        fname = prepareLoading(name, path, extension=extension)

        if self.version == 3:
            self.__parse3(fname, validate=validate)
        else:
            raise ValueError("Requested version not supported!")

    def write(self, name, path, extension="gff", validate=True):
        fname = prepareSaving(name, path, extension=extension)

        if self.version == 3:
            self.__write3(fname, validate=validate)
        else:
            raise ValueError("Requested version not supported!")

    def __parse3(self, fname, validate):
        """
        parse gff version 3 file.
        """
        self._header_region = {}

        splitter = '\t'
        counter = 0
        with open(fname, 'r') as file:
            for line in file:
                counter += 1

                # handel header lines
                if line[0] in self._header_char:

                    if "version" in line:
                        continue

                    if "##sequence-region" in line:
                        split = line.find(' ')
                        if split == -1:
                            raise IOError(
                                "Invalid GFF file! No 'seqid' region header in line {c}".format(
                                    c=counter))
                        key = line[split:]
                        split = key.find(' ')
                        if split == -1:
                            raise IOError(
                                "Invalid GFF file! No 'seqid' region header in line {c}".format(
                                    c=counter))
                        key = key[0:split]
                        key = key.lstrip()
                        self._header_region[key] = line
                    else:
                        self._header_general.append(line)
                    continue

                # find reference_sequence
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid GFF file! No 'seqid' entry in line {c}".format(
                            c=counter))
                seqid = line[0:split]
                line = line[split + 1:]
                line = line.lstrip()

                # find source
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid GFF file! No 'source' entry in line {c}".format(
                            c=counter))
                source = line[0:split]
                line = line[split + 1:]
                line = line.lstrip()

                # find method
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid GFF file! No 'type' entry in line {c}".format(
                            c=counter))
                type = line[0:split]
                line = line[split + 1:]
                line = line.lstrip()

                # find start
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid GFF file! No 'start' entry in line {c}".format(
                            c=counter))
                start = int(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find stop
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid GFF file! No 'end' entry in line {c}".format(
                            c=counter))
                end = int(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find score
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid GFF file! No 'score' entry in line {c}".format(
                            c=counter))
                if line[0:split] == '.':
                    score = None
                else:
                    score = float(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find strand
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid GFF file! No 'strand' entry in line {c}".format(
                            c=counter))
                strand = line[0:split]
                line = line[split + 1:]
                line = line.lstrip(splitter)

                # find phase
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid GFF file! No 'phase' entry in line {c}".format(
                            c=counter))
                if line[0:split] == '.':
                    phase = None
                else:
                    phase = int(line[0:split])
                line = line[split + 1:]

                # remove remaining tab and line end character
                line = line.lstrip()
                line = line.rstrip()

                entry = GFF3Entry(seqid, source, type, start, end, score=score,
                                  strand=strand, phase=phase)

                gsplitter = ';'
                separator = '='
                attributes = {}
                while len(line) > 0:

                    # extract  and value
                    split = line.find(gsplitter)
                    if split == -1:
                        split = len(line)

                    keyvalue = line[0:split]
                    key, _, value = keyvalue.partition(separator)

                    attributes[key] = value
                    line = line[split + 1:]
                    line = line.lstrip(gsplitter)
                    line = line.rstrip(gsplitter)
                    line = line.lstrip()
                    line = line.rstrip()
                if len(attributes) > 0:
                    entry.attributes = attributes
                if validate:
                    if not entry.valid:
                        raise IOError(
                            "Invalid GFF file! Entry in line {c} is {e}".format(
                                c=counter, e=str(entry)))

                # check if key is already there
                self.append(entry)

    def __write3(self, fname, validate):
        with open(fname, 'w') as file:
            head = "##gff-version 3 \n"
            if self._header_general[0] != head:
                self._header_general.insert(0, head)

            # write header first
            for item in self._header_general:
                file.write(item)

            # write data
            for key in self._keys:
                if key in self._header_region:
                    for item in self._header_region[key]:
                        file.write(item)

                for item in self._data[key]:
                    if validate:
                        if not item.valid:
                            raise ValueError(
                                "Can not write file due to invalid GFF entry: " + str(
                                    item))
                        file.write(str(item))




#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Functions and classes for handling next generation sequencing data.
:author: Manuel Tuschen
:license: FreeBSD

License
----------
Copyright (c) 2016, Manuel Tuschen
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


def selectGenes(df, selection):
    """
    Remove all rows (gene_ids) from a data frame which can not be found in
    selection.

    Parameters
    ---------
    df : DataFrame
        The data frame to work with.
    selection : set
        All ids to keep.

    Return
    ------
    df : DataFrame
        The cleand data frame.
    """
    selection = set(selection)

    for id in df.index:
        if id not in selection:
            df = df.drop(id)
    return df

def removeUnexpr(df):

    # we remove 0 fpkms first

    df = df[df.FPKM != 0]
    df = df[df.TPM != 0]
    return df

def keepHighExpr(df, n):
    df.sort_values(by="FPKM",ascending=0)

    return df.ilic([range(0,n)])

def keepVarExpr(df, n):
    pass

def meanVArExpr():
    pass


def selectBound(df, scores, n):


    bound = []
    unbound = []
    nb = 0
    nu = 0

    # go through all gene ids in the given order
    for id in df.index:
        #create list of box_ids
        box_id = ["{id}_{i}".format(id=id, i=i) for i in range(0,60)]
        box_id = set(box_id)

        # extract scores with box_id
        tmp = scores[box_id]

        # lets see if there is at least one box with a positive score
        tmp = tmp > 0
        if tmp.any(0).any(0):
            bound.append(id)
            nb += 1
        else:
            unbound.append(id)
            nu += 1






import datetime
from copy import deepcopy

import numpy as np
import pandas as pd

from GFF import GFF3entry, GFF


def convertGFFtoMIL(scoreGFFs, labelGFF, name, path):

    mil = milData(name)

    ## we have a list of GFFfiles where we can extract scores from
    ## GFF files are assumed to have the same number of data points


    ## load data if not already done
    for gff in scoreGFFs:
        if gff.data is None:
            gff.parse()
    if labelGFF.data is None:
        labelGFF.parse()

    ## check if all gff files have same number of data points
    N_x = len(labelGFF.data)
    for gff in scoreGFFs:
        if len(gff.data) != N_x:
            raise ValueError("All GFF files need same amount of data")


    ## go through all instances
    for i in range(N_x):
        ## collect scores from all GFF files

        scores = np.zeros((len(scoreGFFs)))
        ID = labelGFF.data[i].attributes["ID"]
        for g, gff in enumerate(scoreGFFs):
            if gff.data[i].attributes["ID"] != ID:
                raise ValueError("ID not equal for GFF files")
            scores[g] = gff.data[i].scores
        label = 0
        if labelGFF.data[i].scores > 0:
            label = 1

        mil.add_x(ID, scores, label)

    mil.del_singleEntires()
    mil.scale()
    mil.save(path)



#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Functions and classes for handling next generation sequencing data.
:author: Manuel Tuschen
:license: FreeBSD

License
----------
Copyright (c) 2016, Manuel Tuschen
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


from __future__ import division, absolute_import, unicode_literals, print_function

from .Formats import BED, BEDEntry




def sliding_box(start, stop, size, step=1):
    """
    Find sliding boxes of region

    Parameters:
    -----------
    start: int
        0 based start coordinate of the region.
    stop : int
        0 based stop coordinate of the region, i.e. stop is not part of the
        region anymore;
    size : int
        The size of the box regions
    step : int, optional
        The step size of the sliding box.

    Return
    ------
    regions : list
        List of tuples (start,stop) with the extracted regions.
    """
    # we need to truncate if step is no multiple of stop-start
    if (stop - start) / step != 0:
        overhead = (stop-start) % step

        if overhead % 2 == 0:
            start += overhead // 2
            stop -= overhead // 2
        else:
            start += overhead // 2
            stop -= overhead // 2 + 1
    regions = []
    for i in range(0,(stop - start),step):
        regions.append((i+start, i+start+size))
    return regions


def GFF_sliding_boxes(inGFF, name, path, size, step):
    """
    Read from inGFF file and create sliding boxes for each region.
    Score regions according to the scores in BedGraph file.
    Write all results into GFFout.

    Parameters:
    ----------
    inGFF : GFF
        The GFF file with the original regions. An ID attribute is mandatory.
    name : string
        The name of the output file.
    path : string
        The path of the output file
    size : int
        The size of the box regions
    step : int
        The step size of the sliding box.
    """
    if len(inGFF) == 0:
        raise ValueError("No data in input GFF")
    BEDout = BED()

    # parse through all gene regions
    for gff in inGFF:
        regions = sliding_box(gff.start-1, gff.end, size, step)
        ID = gff.attributes["ID"]
        for i,region in enumerate(regions):
            chrom = gff.seqid
            chromStart = region[0]
            chromEnd = region[1]
            name = "{id}_{i}".format(id=ID, i=i)

            entry = BEDEntry(chrom, chromStart, chromEnd, name=name)
            BEDout.append(entry)
    BEDout.write(name, path)





#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Functions and classes for handling next generation sequencing data.
:author: Manuel Tuschen
:license: FreeBSD

License
----------
Copyright (c) 2016, Manuel Tuschen
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


from miscpy import prepareSaving
from miscpy.DataStructure import hstr


class Data:
    def __init__(self, data_type):
        self._header_char = ['#']

        self._data = None
        self._data_type = data_type
        self._extension = ""

        self._len = 0
        self._keys = []

    def __len__(self):
        return self._len

    def __getitem__(self, key):
        try:
            chrom = hstr(key[0])
            n = key[1]
            return self._data[chrom][n]
        except:
            raise KeyError("Key {k}, not found for BED".format(k=key))

    def __setitem__(self, key, value):
        if not isinstance(value, self._data_type):
            raise ValueError(
                "Value must be of type t!".format(t=self._data_type))
        try:
            if self._data is None:
                self._data = {}
            chrom = hstr(key[0])
            n = key[1]
            self._data[chrom][n] = value
            self._len += 1
        except:
            raise KeyError("Key {k}, not found!".format(k=key))

    def __delitem__(self, key):
        try:
            chrom = hstr(key[0])
            n = key[1]
            del self._data[chrom][n]
            if len(self._data[chrom]) == 0:
                self._keys.remove(chrom)
                del self._data[chrom]
            self._len -= 1
        except:
            raise KeyError("Key {k}, not found!".format(k=key))

    def __iter__(self):
        self.__iter = (0, 0)
        return self

    def __next__(self):
        # see if there is another item for the current chrom
        if self.__iter[1] < len(self._data[self._keys[self.__iter[0]]]) - 1:
            pre_iter = (self.__iter[0], self.__iter[1])
            self.__iter = (self.__iter[0], self.__iter[1] + 1)
            return self._data[self._keys[pre_iter[0]]][pre_iter[1]]
        # see if there is another chrom left
        elif self.__iter[0] < len(self._keys) - 1:
            pre_iter = (self.__iter[0], self.__iter[1])
            self.__iter = (self.__iter[0] + 1, 1)
            return self._data[self._keys[pre_iter[0]]][pre_iter[1]]
        # otherwise we are done
        else:
            raise StopIteration

    def append(self, value):
        if self._data is None:
            self._data = {}
        key = hstr(value.chrom)
        if not isinstance(value, self._data_type):
            raise ValueError(
                "Value must be of type {t}!".format(t=self._data_type))
        if key not in self._data:
            self._keys.append(key)
            self._data[key] = []
        self._data[key].append(value)
        self._len += 1

    def sort(self):
        self._keys = sorted(self._keys)
        for key in self._keys:
            self._data[key] = sorted(self._data[key])

    def write(self, name, path, extension="", validate=True):
        if extension == "":
            extension = self._extension
        fname = prepareSaving(name, path, extension=extension)

        with open(fname, 'w') as file:
            # write data
            for key in self._keys:
                for item in self._data[key]:
                    if validate:
                        if not item.valid:
                            raise ValueError(
                                "Can not write file due to invalid data entry: " + str(
                                    item))
                        file.write(str(item))



#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Functions and classes for handling next generation sequencing data.
:author: Manuel Tuschen
:license: FreeBSD

License
----------
Copyright (c) 2016, Manuel Tuschen
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import division, absolute_import, unicode_literals, \
    print_function
import re

from miscpy import prepareLoading, prepareSaving

from ._ngsData import Data


__all__ = ["BEDEntry", "BED"]


class BED3Entry:
    """
    BED 3 format provides the basic underlying format for the UCSC genome
    browser: https://genome.ucsc.edu/FAQ/FAQformat.html#format1
    It is called BED 3 because it provides 3 entries which are mandatory and
    must be present in all .bed files. Apart from BED 3 there exists several
    derived formats which add additional entries.

 Attributes
    ----------
    chrom : basestring
        The name of the chromosome (e.g. chr3, chrY, chr2_random) or scaffold
        (e.g. scaffold10671)
    chromStart : int
       The starting position of the feature in the chromosome or scaffold. The
       first base in a chromosome is numbered 0.
    chromEnd : int
        The ending position of the feature in the chromosome or scaffold. The
        chromEnd base is not included in the display of the feature. For
        example, the first 100 bases of a chromosome are defined as
        chromStart=0, chromEnd=100, and span the bases numbered 0-99.
    """

    def __init__(self, chrom, chromStart, chromEnd):
        self._sep = "\t"
        self._newline = "\n"
        self.chrom = chrom
        self.chromStart = chromStart
        self.chromEnd = chromEnd

    def __getattr__(self, valid):
        return self._valid()

    def __str__(self):
        sep = self._sep
        newline = self._newline

        chrom = str(self.chrom)
        chromStart = str(self.chromStart)
        chromEnd = str(self.chromEnd)
        return chrom + sep + chromStart + sep + chromEnd + newline

    def __lt__(self, other):
        """
        One BED 3 is regarded larger only if the chromosome is larger or equal
        and the start position is truly larger.
        """
        if self.chrom <= other.chrom and self.chromStart < other.chromStart:
            return True
        else:
            return False

    def __le__(self, other):
        return not other < self

    def __gt__(self, other):
        return other < self

    def __ge__(self, other):
        return not self < other

    def __eq__(self, other):
        """
        We consider equality only if all entries are equal.
        """
        if self.chrom == other.chrom and self.chromStart == other.chromStart and self.chromEnd == other.chromEnd:
            return True
        else:
            return False

    def __neg__(self, other):
        return not self == other

    def _valid(self):
        pattern = re.compile(".")  # we can have an arbitrary string
        # check chrom
        if not pattern.match(self.chrom):
            return False
        # check start
        if not isinstance(self.chromStart, int):
            return False
        if self.chromStart < 0:
            return False
        # check end
        if not isinstance(self.chromEnd, int):
            return False
        if self.chromEnd <= self.chromStart:
            return False
        return True


class BED6Entry(BED3Entry):
    """
    BED 6 format provides some extensions to the BED 3 formatt:
    https://genome.ucsc.edu/FAQ/FAQformat.html#format1
    It is called BED 6 because it provides 3 more fields compared to BED 3.

 Attributes
    ----------
    chrom : basestring
        The name of the chromosome (e.g. chr3, chrY, chr2_random) or scaffold
        (e.g. scaffold10671)
    chromStart : int
       The starting position of the feature in the chromosome or scaffold. The
       first base in a chromosome is numbered 0.
    chromEnd : int
        The ending position of the feature in the chromosome or scaffold. The
        chromEnd base is not included in the display of the feature. For
        example, the first 100 bases of a chromosome are defined as
        chromStart=0, chromEnd=100, and span the bases numbered 0-99.
    name : basestring, optional
        Defines the name of the BED line. This label is displayed to the left
        of the BED line in the Genome Browser window when the track is open to
        full display mode or directly to the left of the item in pack mode.
    score : float, optional
        A score between 0 and 1000. If the track line useScore attribute is set
        to 1 for this annotation data set, the score value will determine the
        level of gray in which this feature is displayed (higher numbers =
         darker gray).
    strand : char, optional
        Defines the strand - either '+' or '-'.
    """

    def __init__(self, chrom, chromStart, chromEnd, score, name=None,
                 strand=None):

        super().__init__(chrom, chromStart, chromEnd)
        self.score = score
        self.name = name
        self.strand = strand

    def __str__(self):
        sep = self._sep
        newline = self._newline

        score = str(self.chrom)
        if self.name is None:
            name = "."
        else:
            name = str(self.name)
        if self.strand is None:
            strand = "."
        else:
            stand = str(self.strand)
        return super().__str__().lstrip() + sep + name + sep + score + sep + strand + newline

    def __eq__(self, other):
        """
        We consider equality only if all entries are equal.
        """
        if super().__eq__(
                other) and self.name == other.name and self.score == other.score and self.strand == other.strand:
            return True
        else:
            return False

    def _valid(self):
        pattern = re.compile(".")  # we can have an arbitrary string
        pattern_strand = re.compile("[+-.]")

        if not super().__valid():
            return False

        # check name
        if not (self.name is None or pattern.match(self.name)):
            return False
        # check score
        if not isinstance(self.signalValue, int):
            return False
        # check strand
        if not (self.strand is None or pattern_strand.match(self.strand)):
            return False
        return True


class BEDEntry(BED3Entry):
    """
    Arbritary BED format as specified in :
    https://genome.ucsc.edu/FAQ/FAQformat.html#format1

    Attributes
    ----------
    chrom : basestring
        The name of the chromosome (e.g. chr3, chrY, chr2_random) or scaffold
        (e.g. scaffold10671)
    chromStart : int
       The starting position of the feature in the chromosome or scaffold. The
       first base in a chromosome is numbered 0.
    chromEnd : int
        The ending position of the feature in the chromosome or scaffold. The
        chromEnd base is not included in the display of the feature. For
        example, the first 100 bases of a chromosome are defined as
        chromStart=0, chromEnd=100, and span the bases numbered 0-99.
    name : basestring, optional
        Defines the name of the BED line. This label is displayed to the left
        of the BED line in the Genome Browser window when the track is open to
        full display mode or directly to the left of the item in pack mode.
    score : float, optional
        A score between 0 and 1000. If the track line useScore attribute is set
        to 1 for this annotation data set, the score value will determine the
        level of gray in which this feature is displayed (higher numbers =
         darker gray).
    strand : char, optional
        Defines the strand - either '+' or '-'.
    thickStart : int, optional
        The starting position at which the feature is drawn thickly (for
        example, the start codon in gene displays). When there is no thick part,
        thickStart and thickEnd are usually set to the chromStart position.
    thickEnd : int, optional
        The ending position at which the feature is drawn thickly (for example,
        the stop codon in gene displays).
    itemRgb : tuble, optional
        An RGB value of the form R,G,B (e.g. 255,0,0). If the track line
        itemRgb attribute is set to "On", this RBG value will determine the
        display color of the data contained in this BED line. NOTE: It is
        recommended that a simple color scheme (eight colors or less) be used
        with this attribute to avoid overwhelming the color resources of the
        Genome Browser and your Internet browser.
    blockCount : int
        The number of blocks (exons) in the BED line.
    blockSizes : list
        A comma-separated list of the block sizes. The number of items in this
        list should correspond to blockCount.
    blockStarts : list
        A comma-separated list of block starts. All of the blockStart positions
        should be calculated relative to chromStart. The number of items in
        this list should correspond to blockCount.
    valid : bool
        To check if the entry is a valid one.

    Methods
    -------
    """

    def __init__(self, chrom, chromStart, chromEnd, name=None, score=None,
                 strand=None, thickStart=None, thickEnd=None, itemRbg=None,
                 blockCount=None, blockSize=None, blockStarts=None):
        super().__init__(chrom, chromStart, chromEnd)
        self.name = name
        self.score = score
        self.strand = strand
        self.thickStart = thickStart
        self.thickEnd = thickEnd
        self.itemRgb = itemRbg
        self.blockCount = blockCount
        self.blockSizes = blockSize
        self.blockStarts = blockStarts

    def __str__(self):
        sep = self._sep
        gsep = ','
        newline = self._newline

        if self.name is None:
            name = ""
        else:
            name = str(self.name)
        if self.score is None:
            score = ""
        else:
            score = str(self.score)
        if self.strand is None:
            strand = ""
        else:
            strand = str(self.strand)
        if self.thickStart is None:
            thickStart = ""
        else:
            thickStart = str(self.thickStart)
        if self.thickEnd is None:
            thickEnd = ""
        else:
            thickEnd = str(self.thickEnd)
        if self.itemRgb is None:
            itemRgb = ""
        else:
            itemRgb = ""
            for item in self.itemRgb:
                itemRgb + str(item) + gsep
            itemRgb = itemRgb.rstrip(gsep)
        if self.blockCount is None:
            blockCount = ""
        else:
            blockCount = str(self.blockCount)
        if self.blockSizes is None:
            blockSizes = ""
        else:
            blockSizes = ""
            for item in self.blockSizes:
                blockSizes + str(item) + gsep
            blockSizes = blockSizes.rstrip(gsep)
        if self.blockStarts is None:
            blockStarts = ""
        else:
            blockStarts = ""
            for item in self.blockStarts:
                blockStarts + str(item) + gsep
            blockStarts = blockStarts.rstrip(gsep)

        val = super().__str__().rstrip() + sep + name + sep + score + sep + strand + sep + thickStart + sep + thickEnd + sep + itemRgb + sep + blockCount + sep + blockSizes + sep + blockStarts
        val = val.rstrip()
        return val + newline

    def __eq__(self, other):

        if super().__eq__(
                other) and self.name == other.naem and self.score == other.score and self.strand == other.strand and self.thickStart == other.thickStart and self.thickEnd == other.thickEnd and self.itemRgb == other.itemRgb and self.blockCount == other.blockCount and self.blockSizes == other.blochSizes and self.blockStarts == other.blockStarts:
            return True
        else:
            return False

    def _valid(self):
        pattern = re.compile(".")
        strand_pattern = re.compile("[+-]]")

        if not super()._valid():
            return False

        # check name:
        if self.name is not None:
            if not pattern.match(self.name):
                return False
        else:
            if not (
                                            self.score is None and self.strand is None and self.thickStart is None and self.thickEnd is None and self.itemRgb is None and self.blockCount is None and self.blockSizes is None and self.blockStarts is None):
                return False

        # check score
        if self.score is not None:
            if not isinstance(self.score, float):
                return False
        else:
            if not (
                                        self.strand is None and self.thickStart is None and self.thickEnd is None and self.itemRgb is None and self.blockCount is None and self.blockSizes is None and self.blockStarts is None):
                return False

        # check strand
        if self.strand is not None:
            if not strand_pattern.match(self.strand):
                return False
        else:
            if not (
                                    self.thickStart is None and self.thickEnd is None and self.itemRgb is None and self.blockCount is None and self.blockSizes is None and self.blockStarts is None):
                return False

        # check thickStart:
        if self.thickStart is not None:
            if not isinstance(self.thickStart, int):
                return False
        else:
            if not (
                                self.thickEnd is None and self.itemRgb is None and self.blockCount is None and self.blockSizes is None and self.blockStarts is None):
                return False

        # check thickEnd:
        if self.thickEnd is not None:
            if not isinstance(self.thickEnd, int):
                return False
        else:
            if not (
                            self.itemRgb is None and self.blockCount is None and self.blockSizes is None and self.blockStarts is None):
                return False

        # check itemRgb:
        if self.itemRgb is not None:
            for item in self.itemRgb:
                if item < 0 or item > 255 and not isinstance(item, int):
                    return False
        else:
            if not (
                        self.blockCount is None and self.blockSizes is None and self.blockStarts is None):
                return False

        # check blockCount:
        if self.blockCount is not None:
            if self.blockSizes is None or self.blockStarts is None:
                return False
            if not isinstance(self.blockCount, int):
                return False
        else:
            if not (self.blockSizes is None and self.blockStarts is None):
                return False

        # check blockSizes:
        if self.blockSizes is not None:
            if len(self.blockSizes) != self.blockCount:
                return False
            for item in self.blockSizes:
                if item < 0 and not isinstance(item, int):
                    return False
        else:
            if not (self.blockSizes is None and self.blockStarts is None):
                return False

        # check blockStarts:
        if self.blockStarts is not None:
            if len(self.blockStarts) != self.blockCount:
                return False
            if self.blockStarts[0] != 0:
                return False
            if self.blockSizes[-1] + self.blockStarts[-1] != self.chromEnd:
                return False
            for item in self.blockStarts:
                if item < 0 and not isinstance(item, int):
                    return False
        return True


class BED(Data):
    """
    A BED object is designed to parse data from an BED file or write its
    data to an BED file. BED supports iterator and mutable sequence protocols.
    Items can be accessed by an tuple key (chr, n), i.e. the nth entry of chr.

    Attributes
    ----------

    Methods
    -------
    append(value) :
        Append another BEDEntry.
    sort() :
        Sort entries;
    parse(name, path, extension="bed", validate=True) :
        Parse data from path/name.extension and store them in the BED object.
    write(name, path, extension="bed", validate=True) :
        Write data to path/name.extension
    """

    def __init__(self):
        super().__init__(BEDEntry)
        self._extension = "bed"

    def parse(self, name, path, extension="bed", validate=True):
        """
        parse BED file.
        """
        fname = prepareLoading(name, path, extension=extension)

        splitter = '\t'
        counter = 0
        with open(fname, 'r') as file:
            self._data = {}

            for line in file:
                counter += 1
                # handel header files
                if line[0] in self._header_char:
                    continue

                # find chrom
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BED file! No 'chrom' entry in line {c}".format(
                            c=counter))
                chrom = line[0:split]
                line = line[split + 1:]
                line = line.lstrip()

                # find chromStart
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BED file! No 'chromStart' entry in line {c}".format(
                            c=counter))
                chromStart = int(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find chromEnd
                split = line.find(splitter)
                if split == -1:
                    chromEnd = int(line[0:])
                else:
                    chromEnd = int(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                entry = BEDEntry(chrom, chromStart, chromEnd)
                # look for optional fields:

                if len(line) > 0:
                    # find name
                    split = line.find(splitter)
                    if split == -1:
                        entry.name = line[0:]
                    else:
                        entry.name = line[0:split]
                    line = line[split + 1:]
                    line = line.lstrip()

                if len(line) > 0:
                    # find score
                    split = line.find(splitter)
                    if split == -1:
                        entry.score = float(line[0:])
                    else:
                        entry.score = float(line[0:split])
                    line = line[split + 1:]
                    line = line.lstrip()

                if len(line) > 0:
                    # find strand
                    split = line.find(splitter)
                    if split == -1:
                        entry.strand = line[0:split]
                    else:
                        entry.strand = line[0:split]
                    line = line[split + 1:]
                    line = line.lstrip()

                if len(line) > 0:
                    # find thickStart
                    split = line.find(splitter)
                    if split == -1:
                        entry.thickStart = int(line[0:])
                    else:
                        entry.thickStart = int(line[0:split])
                    line = line[split + 1:]
                    line = line.lstrip()

                if len(line) > 0:
                    # find thickEnd
                    split = line.find(splitter)
                    if split == -1:
                        entry.thickEnd = int(line[0:])
                    else:
                        entry.thickEnd = int(line[0:split])
                    line = line[split + 1:]
                    line = line.lstrip()

                if len(line) > 0:
                    # find itemRgb
                    entry.itemRgb = []
                    split = line.find(splitter)
                    if split == -1:
                        value = line[0:]
                    else:
                        value = line[0:split]
                    while len(value) > 0:
                        sep = value.find(',')
                        if sep == -1:
                            entry.itemRgb.append(int(value[0:]))
                            break
                        else:
                            entry.itemRgb.append(int(value[0:sep]))
                        value = value[sep + 1:]
                        value = value.lstrip()

                    line = line[split + 1:]
                    line = line.lstrip()

                if len(line) > 0:
                    # find blockCount
                    split = line.find(splitter)
                    if split == -1:
                        entry.blockCount = int(line[0:])
                    else:
                        entry.blockCount = int(line[0:split])
                    line = line[split + 1:]
                    line = line.lstrip()

                if len(line) > 0:
                    # find blockSizes
                    entry.blockSizes = []
                    split = line.find(splitter)
                    if split == -1:
                        value = line[0:]
                    else:
                        value = line[0:split]

                    while len(value) > 0:
                        sep = value.find(',')
                        if sep == -1:
                            entry.blockSizes.append(int(value[0:]))
                            break
                        else:
                            entry.blockSizes.append(int(value[0:sep]))
                        value = value[sep + 1:]
                        value = value.lstrip()
                    line = line[split + 1:]
                    line = line.lstrip()

                if len(line) > 0:
                    # find blockStarts
                    entry.blockStarts = []
                    split = line.find(splitter)
                    if split == -1:
                        value = line[0:]
                    else:
                        value = line[0:split]
                    while len(value) > 0:
                        sep = value.find(',')
                        if sep == -1:
                            entry.blockStarts.append(int(value[0:]))
                            break
                        else:
                            entry.blockStarts.append(int(value[0:sep]))
                        value = value[sep + 1:]
                        value = value.lstrip()

                if validate:
                    if not entry.valid:
                        raise IOError(
                            "Invalid BED file! Entry in line {c} is {e}".format(
                                c=counter, e=str(entry)))

                self.append(entry)




"""
Functions and classes for handling next generation sequencing data.
:author: Manuel Tuschen
:license: FreeBSD

License
----------
Copyright (c) 2016, Manuel Tuschen
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import division, absolute_import, unicode_literals, \
    print_function

from miscpy import prepareLoading, prepareSaving

from ._ngsData import Data
from .ngsBED import BED3Entry


__all__ = ["BedGraphEntry", "BedGraph"]


class BedGraphEntry(BED3Entry):
    """
    BedGraph entry.
    https://genome.ucsc.edu/FAQ/FAQformat.html#format1

    Notes on comparing GFF entries.
    Entries are considered equal only if all entries are equal.
    One entry is regarded as larger only if the sequence id is larger or equal
    and the start position is truly larger. One entry is regarded as
    larger-equal only if the sequence id is larger or equal and the start
    position is larger or equal. The same rules apply for smaller.

    Attributes
    ----------
    chrom : basestring
        The name of the chromosome (e.g. chr3, chrY, chr2_random) or scaffold
        (e.g. scaffold10671)
    chromStart : int
        The starting position of the feature in the chromosome or scaffold. The
        first base in a chromosome is numbered 0.
    chromEnd : int
        The ending position of the feature in the chromosome or scaffold.
        The chromEnd base is not included in the display of the feature. For
        example, the first 100 bases of a chromosome are defined as
        chromStart=0, chromEnd=100, and span the bases numbered 0-99.
    score : float
        A score between 0 and 1000. If the track line useScore attribute is set
        to 1 for this annotation data set, the score value will determine the
        level of gray in which this feature is displayed (higher numbers =
         darker gray).
    valid : bool
        To check if the entry is a valid one.

    Methods
    -------
    """

    def __init__(self, chrom, chromStart, chromEnd, score):
        super().__init__(chrom, chromStart, chromEnd)
        self.score = score

    def __str__(self):
        sep = self._sep
        newline = self._newline
        score = str(self.score)

        return super().__str__().rstrip() + sep + score + newline

    def __eq__(self, other):

        if super().__eq__(other) and self.score == other.score:
            return True
        else:
            return False

    def _valid(self):

        if not super()._valid():
            return False
        # check score
        if not isinstance(self.score, float):
            return False

        return True


class BedGraph(Data):
    """
    A BedGraph object is designed to parse data from an BedGraph file
    or write its data to an BedGraph file. BedGraph supports iterator and
    mutable sequence protocols. Items can be accessed by an tuple key (chr, n),
    i.e. the nth entry of chr.

    Attributes
    ----------

    Methods
    -------
    append(value) :
        Append another BEDentry.
    sort() :
        Sort entries;
    parse(name, path, extension="bedGraph", validate=True) :
        Parse data from path/name.extension and store them in the BED object.
    write(name, path, extension="bedGraph", validate=True) :
        Write data to path/name.extension
    """

    def __init__(self):
        super().__init__(BedGraphEntry)
        self._extension = "bedGraph"

    def parse(self, name, path, extension="bedGraph", validate=True):
        """
        parse BedGraph file.
        """

        fname = prepareLoading(name, path, extension=extension)
        with open(fname, 'r') as file:
            splitter = '\t'
            counter = 0
            for line in file:
                counter += 1

                # skip header lines
                if line[0] in self._header_char:
                    continue

                # parse data
                # find chrom
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BedGraph file! No 'chrom' entry in line {c}".format(
                            c=counter))
                chrom = line[0:split]
                line = line[split + 1:]
                line = line.lstrip()

                # find start
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BedGraph file! No 'chromStart' entry in line {c}".format(
                            c=counter))
                chromStart = int(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find end
                split = line.find(splitter)
                if split == -1:
                    raise IOError(
                        "Invalid BedGraph file! No 'chromEnd' entry in line {c}".format(
                            c=counter))
                chromEnd = int(line[0:split])
                line = line[split + 1:]
                line = line.lstrip()

                # find score
                line = line.rstrip()
                score = float(line)

                entry = BedGraphEntry(chrom, chromStart, chromEnd, score)
                if validate:
                    if not entry.valid:
                        raise IOError(
                            "Invalid BedGraph file! Entry in line {c} is {e}".format(
                                c=counter, e=str(entry)))
                self.append(entry)



#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Functions and classes for handling next generation sequencing data.
:author: Manuel Tuschen
:license: FreeBSD

License
----------
Copyright (c) 2016, Manuel Tuschen
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


from __future__ import division, absolute_import, unicode_literals, print_function

import numpy as np
import pandas as pd

from miscpy import prepareSaving


def scoreRegion(inGFF, inBedGraph, name, path):
    """
    Score a genomic regions specified in an GFF by averaging over a scores from
    a BedGraph object. Scores are written to a .tsv .

    Parmeters
    ---------
    inGFF : ngsGFF
        The GFF object with genomic regions to score.
    inBedGraph : ngsBedGraph
        The BedGraph object with scores for each region.
    name : basestring
        The filename for the output file.
    path : basestring
        The path to write the  output to.
    """
    if len(inGFF) == 0:
        raise ValueError("GFF does not contain any data!")
    if len(inBedGraph) == 0:
        raise ValueError("BedGraph does not contain any data")

    inGFF.sort()
    inBedGraph.sort()

    chrom = []
    start = []
    end = []
    coverage = []
    sum = []
    mean = []
    mean0 = []
    min = []
    max = []

    header = ["chrom", "start", "end", "coverage", "sum", "mean", "mean0", "min", "max"]
    id = []

    # we need to loop over all gff regions
    for gff in inGFF:
        cover = 0
        mi = np.inf
        ma = -np.inf
        score = 0
        chr = gff.seqid
        gff_start = gff.start - 1 # gff format is 1 based and end inclusive
        # and the we need to loop over all BedGraph regions but only for the
        # correct chrom
        for bedgr in inBedGraph.__data[chr]:

            # lets check if we have coverage at
            # gff:   ==============================
            # bg:  -                                -
            if bedgr.chromEnd <= gff_start:
                continue
            if bedgr.chromStart >= gff.end:
                continue

            # complete coverage if gff fragment is smaller:
            # gff:   ==============================
            # bg:    -        -                   -
            # bg:    ------------------------------

            if bedgr.chromEnd <= gff_start and bedgr.chromEnd <= gff.end:
                coverage += bedgr.chromeEnd - bedgr.chromStart
                score += bedgr.score
                mi = min(mi, bedgr.score)
                ma = max(ma, bedgr.score)
                continue

            # partial coverage at the beginning
            # gff:   ==============================
            # bg:   --
            # bg:   -------------------------------
            # bg:   -------------------------------------
            if bedgr.chromStart < gff_start:
                len = min(gff.end, bedgr.chromEnd) - gff_start
                coverage += len
                score += bedgr.score / (bedgr.chromEnd - bedgr.chromStart) * len
                mi = min(mi, bedgr.score)
                ma = max(ma, bedgr.score)
                continue

            # partial coverage at the end
            # gff:   ==============================
            # bg:                                 --
            # bg:    -------------------------------
            if bedgr.chromEnd > gff.end:
                len = bedgr.chromEnd - max(gff_start, bedgr.chromStart)
                coverage += len
                score += bedgr.score / (bedgr.chromEnd - bedgr.chromStart) * len
                mi = min(mi, bedgr.score)
                ma = max(ma, bedgr.score)
                continue

        chrom.append(chr)
        start.append(gff_start)
        end.append(gff.end)
        id.append(gff.attributes["ID"])
        coverage.append(cover)
        sum.append(score)
        mean.append(score / cover)
        mean0.append(score / (gff.end - gff_start))
        min.append(mi)
        max.append(ma)

    # finally we want to store the data as a csv file
    df = pd.DataFrame([chrom, start, end, coverage, sum, mean, mean0, min, max], columns=id)
    df = df.T
    df.columns = header

    fname = prepareSaving(name, path, extension="tsv")
    df.to_csv(fname, sep="\t")


from distutils.core import setup


setup(name='ngspy',
      version='0.1.0',
      description='ngspy collects various tool to deal with next generation sequencing data.',
      url='https://github.com/manuSrep/ngspy.git',
      author='Manuel Tuschen',
      author_email='Manuel_Tuschen@web.de',
      license='FreeBSD License',
      packages=['ngspy', "ngspy/Formats"],
      zip_safe=False)



#!/usr/bin/python
# -*- coding: utf8 -*-

"""
Functions and classes for handling next generation sequencing data.
:author: Manuel Tuschen
:license: FreeBSD

License
----------
Copyright (c) 2016, Manuel Tuschen
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


from __future__ import division, absolute_import, unicode_literals, print_function
import datetime

from .Formats import GFF, GFF3Entry




def extractProm(inGFF, name, path, upstram=499, downstream=100, filter={}):
    """
    Extract promotor regions from GFF file as provided by GENCODE:
    http://www.gencodegenes.org

    Genes can be filtered by GFF fields and attributes. Results are stored
    in a new GFF file.

    Parameters
    ---------
    inGFF : GFF
        The GFF object with a filled dataset. The entries describe the gene
        regions. From these regions the promoter is extracted.
    name : basestring
        The filename of the resulting GFF file.
    path : basestring
        The path where to save the results
    upstream : int
        Determines how many genes upstream of the TSS shall regarded as
        promoter. (in front of the gene)
    downstream : int
        Determines how many genes downstream of the TSS shall regarded as
        promoter. (inside the gene)
    filter : dict
        Regions in the GFF are only considerd if they match the filter entries.
        Keys can be any attribute in the GFFEntry. For most tags it is checked
        if the attribute value is in the filter. For "start" it is checked if
        the value is higher than the filter. The same is for score. For "end"
        it is checked if the value is lower than the filter.
    """
    def check_attr(filter_attr, item_attr):
        for key in filter_attr.keys():
            if not item_attr[key] in filter_attr[key]:
                return False
        return True

    if len(inGFF) == 0:
        raise ValueError("No data in input GFF")

    outGFF = GFF()
    outGFF._header_general.append("## Promotors extracted at: " + str(datetime.date.today()) + "\n")

    for gff in inGFF:
        # filter if item shall be extracted

        # shall we filer for seqid
        if "seqid" in filter:
            if not gff.seqid in filter["seqid"]:
                continue
        if "source" in filter:
            if not gff.source in filter["source"]:
                continue
        if "type" in filter:
            if not gff.type in filter["type"]:
                continue
        if "start" in filter:
            if not gff.start > filter["start"]:
                continue
        if "end" in filter:
            if not gff.end < filter["end"]:
                continue
        if "score" in filter:
            if gff.score < filter["score"]:
                continue
        if "strand" in filter:
            if not gff.strand in filter["strand"]:
                continue
        if "phase" in filter:
            if not gff.phase in filter["phase"]:
                continue
        if "attributes" in filter:
            if not check_attr(filter["attributes"], gff.attributes):
                continue

        seqid = gff.seqid
        source = "ROI"
        type = "promotor"
        strand = gff.strand
        attributes = {}
        attributes["ID"] = gff.attributes["ID"]

        # TSS varies from + and - strands
        if gff.strand == '-':
            start = gff.end - downstream
            end = gff.end + upstram
        else: # = is assumed for '.'
            start = gff.start - upstram
            end = gff.start + downstream

        entry = GFF3Entry(seqid, source, type, start, end, strand=strand, attributes=attributes)
        outGFF.append(entry)

    outGFF.sort()
    outGFF.write(name, path)


"""
Functions and classes for handling next generation sequencing data.
:author: Manuel Tuschen
:license: FreeBSD

License
----------
Copyright (c) 2016, Manuel Tuschen
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import division, absolute_import, unicode_literals, \
    print_function

from miscpy import prepareLoading

from ._ngsData import Data
from .ngsBedGraph import BedGraphEntry


__all__ = ["WIG"]


class WIG(Data):
    """
    A BedGraph object is designed to parse data from an BedGraph file
    or write its data to an BedGraph file. BedGraph supports iterator and
    mutable sequence protocols. Items can be accessed by an tuple key (chr, n),
    i.e. the nth entry of chr.

    Attributes
    ----------

    Methods
    -------
    append(value) :
        Append another BEDentry.
    sort() :
        Sort entries;
    parse(name, path, extension="bedGraph", validate=True) :
        Parse data from path/name.extension and store them in the BED object.
    write(name, path, extension="bedGraph", validate=True) :
        Write data to path/name.extension
    """

    def __init__(self):
        super().__init__(BedGraphEntry)

    def parse(self, name, path, extension="bedGraph", validate=True):
        """
        parse WIG file.
        """

        fname = prepareLoading(name, path, extension=extension)
        with open(fname, 'r') as file:
            splitter = '\t'
            counter = 0
            for line in file:
                counter += 1

                # skip header lines
                if line[0] in self.__header_char:
                    continue

                # parse data
                variable = True
                chrom = None
                start = None
                step = None
                span = 1
                # parse step lines

                if "variableStep" in line:
                    variable = True
                    line = line.lstrip("variableStep")
                    line = line.lstrip()
                    line = line.rstrip()

                    # see if there is a span field
                    split = line.find(splitter)
                    if split == -1:
                        # no span so line should be only chrom field
                        _, _, c = line.partition("=")
                        chrom = c
                    else:
                        tmp = line[0:split]
                        _, _, c = tmp.partition("=")
                        chrom = c
                        line = line[split + 1:]
                        _, _, s = line.partition("=")
                        span = int(s)
                    continue

                if "fixedStep" in line:
                    variable = False
                    line = line.lstrip("fixedStep")
                    line = line.lstrip()
                    line = line.rstrip()
                    # look for the chrom
                    split = line.find(splitter)
                    if split == -1:
                        raise ValueError("No 'chrom' field in WIG file!")
                    else:
                        tmp = line[0:split]
                        _, _, c = tmp.partition("=")
                        chrom = c
                    line = line[split + 1:]
                    line = line.lstrip()
                    # look for start
                    if split == -1:
                        raise ValueError("No 'start' field in WIG file!")
                    else:
                        tmp = line[0:split]
                        _, _, s = tmp.partition("=")
                        start = int(start)
                    line = line[split + 1:]
                    line = line.lstrip()
                    # see if there is a span field
                    split = line.find(splitter)
                    if split == -1:
                        # no span so only step field left
                        _, _, s = line.partition("=")
                        step = int(s)
                    else:
                        tmp = line[0:split]
                        _, _, s = tmp.partition("=")
                        step = int(s)
                        line = line[split + 1:]
                        line = line.lstrip()
                        _, _, s = line.partition("=")
                        span = int(s)
                    continue

                if variable:
                    line = line.lstrip()
                    line = line.rstrip()
                    st, _, sc = line.partition(splitter)
                    entry = BedGraphEntry(chrom, int(st), int(st) + span,
                                          float(sc))
                    if validate:
                        if not entry.valid:
                            raise IOError(
                                "Invalid BedGraph file! Entry in line {c} is {e}".format(
                                    c=counter, e=str(entry)))
                    self.append(entry)
                else:
                    line = line.lstrip()
                    line = line.rstrip()
                    entry = BedGraphEntry(chrom, start, start + span,
                                          float(line))
                    entry = BedGraphEntry(chrom, int(st), int(st) + span,
                                          float(sc))
                    if validate:
                        if not entry.valid:
                            raise IOError(
                                "Invalid BedGraph file! Entry in line {c} is {e}".format(
                                    c=counter, e=str(entry)))
                    self.append(entry)
                    start += step

    def write(self, name, path, extension="", validate=True):
        pass
